from __future__ import absolute_import, division, print_function

from detox.nn.autoencoder import Autoencoder, autoencoder
from detox.nn.wgan import WGAN, wgan

__all__ = [
    # Classes
    'Autoencoder',
    'WGAN',
    # Functionals
    'autoencoder',
    'wgan',
]
